/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved via the world wide web at http://www.erlang.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Original Code is Scott Lystig Fritchie.
 * Portions created by Mr. Fritchie are Copyright 2003, Scott Lystig
 * Fritchie, <slfritchie@snookles.com>. All Rights Reserved.''
 * 
 *     $Id$
 */
#ifndef _DB_JUDY_H
#define _DB_JUDY_H

#define JUDYERROR_NOTEST
#include <Judy.h>
#include "erl_db_util.h" /* DbTerm */

 /* Max size of serialized key and encoded serialized key */
#define	DB_JUDY_BINKEY_MAXSIZE	(4*1024)
#define	DB_JUDY_ENCKEY_MAXSIZE	(DB_JUDY_BINKEY_MAXSIZE + \
				 (DB_JUDY_BINKEY_MAXSIZE/8) + 100/*XXXfudge*/)

typedef struct judy_db_term {
    DbTerm dbterm;                       /* The actual term */
} JudyDbTerm;

/*
** AAAAAARRRRRRRGGGGGGGGGHHHHHHHHH!!!!!!  The DbTable union is abused in
** a way that I wasn't expecting!!!!!!!!!!!!!
**
** THE FIRST MEMBERS OF ANY ETS TABLE STRUCTURE **MUST** BE "owner"
** THROUGH "fixations".  DO NOT MODIFY THEM OR INSERT MEMBERS ANYWHERE
** BEFORE "fixations"!
*/

typedef struct db_table_judy {
    Eterm  owner;             /* Pid of the creator */
    Eterm  the_name;          /* an atom   */
    Eterm  id;                /* atom | integer */
    Uint32 status;            /* bit masks defining type etc */
    int slot;                 /* slot in db_tables */
    int keypos;               /* defaults to 1 */
    int nitems;               /* Total number of items */
    Uint memory;              /* Total memory size */
    /* FWIW, IA32 cache line boundary probably here */
    int kept_items;           /* Always empty for trees */
    Uint megasec,sec,microsec; /* Last fixation time */
    DbFixation *fixations;     /* List of processes who have fixed 
				  the table */
    /* User-servicable parts go below here */
    Pvoid_t root;	      /* The tree root */
    char *binkey;             /* serialized key buffer */
    char *enckey;             /* NUL-terminated key buffer */
    /* FWIW, IA32 cache line boundary probably here */
} DbTableJudy;

/*
** Function prototypes, looks the same (except the suffix) for all 
** table types. The process is always an [in out] parameter.
**
** The functions appear here in order of implementation, in order
** to help keep track of what "ought" to be done next.
*/
void db_initialize_judy(void);
int db_create_judy(Process *p, DbTableJudy *tb /* [in out] */);
void free_judy_table(DbTableJudy *tb /* [in out] */);
int db_put_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		Eterm obj /* [in] */,
		Eterm *ret /* [out] */);
int db_get_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		Eterm key /* [in] */,
		Eterm *ret /* [out] */);
int db_member_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		   Eterm key /* [in] */,
		   Eterm *ret /* [out] */);
int db_first_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		  Eterm *ret /* [out] */);
int db_next_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		 Eterm key /* [in] */,
		 Eterm *ret /* [out] */);
int db_last_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		  Eterm *ret /* [out] */);
int db_prev_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		 Eterm key /* [in] */,
		 Eterm *ret /* [out] */);
int db_erase_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		  Eterm key /* [in] */,
		  Eterm *ret /* [out] */);
int db_update_counter_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			   Eterm key /* [in] */,
			   Eterm incr, /* [in] */
			   int counterpos, /* [in] */
			   Eterm *ret /* [out] */);
/* QQQ Not yet implemented, from here to EOF.... */
int db_get_element_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			Eterm key /* [in] */,
			int ndex, /* [in] */
			Eterm *ret /* [out] */);
int db_erase_object_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			 Eterm object /* [in] */,
			 Eterm *ret /* [out] */);
int db_slot_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		  Eterm slot_term /* [in] */,
		  Eterm *ret /* [out] */);
int db_select_judy(Process *p, DbTableJudy *tb /* [in out] */, 
		   Eterm pattern /* [in] */, 
		   int reversed /* [in] */,
		   Eterm *ret /* [out] */);
int db_select_count_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			 Eterm pattern /* [in] */, 
			 Eterm *ret /* [out] */);
int db_select_chunk_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			 Eterm pattern /* [in] */, 
			 Sint chunk_size /* [in] */,
			 int reversed /* [in] */,
			 Eterm *ret /* [out] */);
int db_select_judy_continue(Process *p, 
			    DbTableJudy *tb /* [in out] */,
			    Eterm continuation /* [in] */,
			    Eterm *ret /* [out] */);
int db_select_count_judy_continue(Process *p, 
				  DbTableJudy *tb /* [in out] */,
				  Eterm continuation /* [in] */,
				  Eterm *ret /* [out] */);
int db_select_delete_judy(Process *p, DbTableJudy *tb /* [in out] */, 
			  Eterm pattern /* [in] */, 
			  Eterm *ret /* [out] */);
int db_select_delete_continue_judy(Process *p, 
				   DbTableJudy *tb /* [in out] */, 
				   Eterm continuation /* [in] */,
				   Eterm *ret);
void db_print_judy(CIO fd /* [in] */, 
		   int show /* [in] */,
		   DbTableJudy *tb /* [in] */);

void erts_db_judy_foreach_offheap(DbTableJudy *tab,
				  void (*)(ErlOffHeap *, void *),
				  void *);

#ifdef HARDDEBUG
void db_check_table_judy(DbTableJudy *tb);
#endif

/*
** XXX XXX XXX This stuff should go in a separate file, but I'm too
** lazy to do it now.
*/

/* B-tree node width constants: each node can store up to MM keys */
#define M	16
#define	MM	(2*M)

typedef struct node {
    int		cnt;
    /*
    ** NOTE!  The DbTerm structure is _variable_length_!!  That's why
    ** is always appears as the last structure member of HashDbTerm et al.
    **
    ** All other ETS table structures only store a single tuple per
    ** {METHOD}DbTerm structure, so they can afford to embed a DbTerm
    ** inside their {METHOD}DbTerm structure.  This permits them to 
    ** call safe_alloc only once, killing two birds with one stone.
    ** We don't have that luxury, alas.  Our consolation is that we
    ** don't have to allocate nearly as many of these as the AVL
    ** code must.
    */
    DbTerm	*key[MM];
    struct node	*ptr[MM+1];
} Node_t;

typedef struct btree_stack {
    Node_t	*node;
    int		idx;
} Btree_Stack_t;

/*
** AAAAAARRRRRRRGGGGGGGGGHHHHHHHHH!!!!!!  The DbTable union is abused in
** a way that I wasn't expecting!!!!!!!!!!!!!
**
** THE FIRST MEMBERS OF ANY ETS TABLE STRUCTURE **MUST** BE "owner"
** THROUGH "fixations".  DO NOT MODIFY THEM OR INSERT MEMBERS ANYWHERE
** BEFORE "fixations"!
*/

typedef struct db_table_btree {
    Eterm  owner;             /* Pid of the creator */
    Eterm  the_name;          /* an atom   */
    Eterm  id;                /* atom | integer */
    Uint32 status;            /* bit masks defining type etc */
    int slot;                 /* slot in db_tables */
    int keypos;               /* defaults to 1 */
    int nitems;               /* Total number of items */
    Uint memory;              /* Total memory size */
    int kept_items;           /* Always empty for trees */
    Uint megasec,sec,microsec;/* Last fixation time */
    DbFixation *fixations;    /* List of processes who have fixed 
				  the table */
    /* User-servicable parts go below here */
    Node_t *root;	      /* The tree root */
    Btree_Stack_t *stack;     /* The first/next stack */
    Uint stack_pos;           /* Current position on stack */
} DbTableBtree;

void db_initialize_btree(void);
int db_create_btree(Process *p, DbTableBtree *tb /* [in out] */);
void free_btree_table(DbTableBtree *tb /* [in out] */);
int db_put_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		Eterm obj /* [in] */,
		Eterm *ret /* [out] */);
int db_get_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		Eterm key /* [in] */,
		Eterm *ret /* [out] */);
int db_member_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		   Eterm key /* [in] */,
		   Eterm *ret /* [out] */);
void db_print_btree(CIO fd /* [in] */, 
		   int show /* [in] */,
		   DbTableBtree *tb /* [in] */);
int db_first_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		  Eterm *ret /* [out] */);
int db_last_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		  Eterm *ret /* [out] */);
int db_next_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		 Eterm key /* [in] */,
		 Eterm *ret /* [out] */);
int db_prev_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		 Eterm key /* [in] */,
		 Eterm *ret /* [out] */);
int db_erase_btree(Process *p, DbTableBtree *tb /* [in out] */, 
		  Eterm key /* [in] */,
		  Eterm *ret /* [out] */);
int db_update_counter_btree(Process *p, DbTableBtree *tb /* [in out] */, 
			   Eterm key /* [in] */,
			   Eterm incr, /* [in] */
			   int counterpos, /* [in] */
			   Eterm *ret /* [out] */);

/* INSERT_* and DELETE_* are magic constants, do not change! */
#define	INSERT_INCOMPLETE      	0
#define	INSERT_OK		1
#define	INSERT_DUP		2

#define	DELETE_UNDERFLOW	0
#define	DELETE_OK		1
#define	DELETE_NOTFOUND		2

#endif /* _DB_JUDY_H */

