%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%% Purpose : Core Erlang syntax trees as records.

%% It would be nice to incorporate some generic functions as well but
%% this could make including this file difficult.

%% Note: the annotation list is *always* the first record field.
%% Thus it is possible to define the macros:
%% -define(get_ann(X), element(2, X)).
%% -define(set_ann(X, Y), setelement(2, X, Y)).

-record(c_int, {anno=[], val}).		% val :: integer()

-record(c_float, {anno=[], val}).	% val :: float()

-record(c_atom, {anno=[], val}).	% val :: atom()

-record(c_char, {anno=[], val}).	% val :: char()

-record(c_string, {anno=[], val}).	% val :: string()

-record(c_nil, {anno=[]}).

-record(c_binary, {anno=[], segs}).	% segs :: [#ce_bin_seg{}]

-record(c_bin_seg, {anno=[],val,	% val :: Tree,
		    size,		% size :: Tree,
		    unit,		% unit :: integer(),
		    type,		% type :: atom(),
		    flags}).		% flags :: [atom()],

-record(c_cons, {anno=[], hd,		% hd :: Tree,
		 tl}).			% tl :: Tree

-record(c_tuple, {anno=[], es}).	% es :: [Tree]

-record(c_var, {anno=[], name}).	% name :: integer() | atom()

-record(c_fname, {anno=[], id,		% id :: atom(),
		  arity}).		% arity :: integer()

-record(c_values, {anno=[], es}).	% es :: [Tree]

-record(c_fun, {anno=[], vars,		% vars :: [Tree],
		body}).			% body :: Tree

-record(c_seq, {anno=[], arg,		% arg :: Tree,
		body}).			% body :: Tree

-record(c_let, {anno=[], vars,		% vars :: [Tree],
		arg,			% arg :: Tree,
		body}).			% body :: Tree

-record(c_letrec, {anno=[], defs,	% defs :: [#ce_def{}],
		   body}).		% body :: Tree

-record(c_def, {anno=[], name,		% name :: Tree,
		val}).			% val :: Tree,

-record(c_case, {anno=[], arg,		% arg :: Tree,
		 clauses}).		% clauses :: [Tree]

-record(c_clause, {anno=[], pats,	% pats :: [Tree],
		   guard,		% guard :: Tree,
		   body}).		% body :: Tree

-record(c_alias, {anno=[], var,		% var :: Tree,
		  pat}).		% pat :: Tree

-record(c_receive, {anno=[], clauses,	% clauses :: [Tree],
		    timeout,		% timeout :: Tree,
		    action}).		% action :: Tree

-record(c_apply, {anno=[], op,		% op :: Tree,
		  args}).		% args :: [Tree]

-record(c_call, {anno=[], module,	% module :: Tree,
		 name,			% name :: Tree,
		 args}).		% args :: [Tree]

-record(c_primop, {anno=[], name,	% name :: Tree,
		   args}).		% args :: [Tree]

-record(c_try, {anno=[], expr,		% expr :: Tree,
		vars,			% vars :: [Tree],
		body}).			% body :: Tree

-record(c_catch, {anno=[], body}).	% body :: Tree

-record(c_module, {anno=[], name,	% name :: Tree,
		   exports,		% exports :: [Tree],
		   attrs,		% attrs :: [#ce_def{}],
		   defs}).		% defs :: [#ce_def{}]
